import angr


class htons(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, to_convert):
        if self.state.arch.memory_endness == "Iend_LE":
            return to_convert[15:0].reversed.zero_extend(len(to_convert) - 16)
        else:
            return to_convert



import ailment

from ...utils.constants import is_alignment_mask


class ArithmeticExpression:
    Add = 0
    Sub = 1
    Or = 2
    And = 4
    RShift = 8
    LShift = 16
    Mul = 32
    Xor = 64

    CONST_TYPES = (int, ailment.expression.Const)

    __slots__ = (
        "op",
        "operands",
    )

    def __init__(self, op, operands):
        self.op = op
        self.operands = operands

    def __repr__(self):
        if self.op == ArithmeticExpression.Add:
            return "%s + %s" % self.operands
        elif self.op == ArithmeticExpression.Sub:
            return "%s - %s" % self.operands
        elif self.op == ArithmeticExpression.And:
            return "%s & %s" % self.operands
        elif self.op == ArithmeticExpression.Or:
            return "%s | %s" % self.operands
        elif self.op == ArithmeticExpression.RShift:
            return "%s >> %s" % self.operands
        elif self.op == ArithmeticExpression.LShift:
            return "%s << %s" % self.operands
        elif self.op == ArithmeticExpression.Mul:
            return "%s * %s" % self.operands
        else:
            return "Unsupported op %s" % self.op

    def __add__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) in ArithmeticExpression.CONST_TYPES:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0] + other,
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        self.operands[1] + other,
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                self,
                other,
            ),
        )

    def __sub__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0] - other,
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        self.operands[1] - other,
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                self,
                other,
            ),
        )

    def __rsub__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) is int:
                return ArithmeticExpression(
                    self.op,
                    other
                    - (
                        self.operands[0],
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        other - self.operands[1],
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                self,
                other,
            ),
        )

    def __and__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0] & other,
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        self.operands[1] & other,
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                self,
                other,
            ),
        )

    def __or__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0] | other,
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        self.operands[1] | other,
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                self,
                other,
            ),
        )

    def __xor__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0] ^ other,
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        self.operands[1] ^ other,
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                self,
                other,
            ),
        )

    def __lshift__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) in ArithmeticExpression.CONST_TYPES:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0] << other,
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        self.operands[1] << other,
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                self,
                other,
            ),
        )

    def __rlshift__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) in ArithmeticExpression.CONST_TYPES:
                return ArithmeticExpression(
                    self.op,
                    (
                        other << self.operands[0],
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        other << self.operands[1],
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                other,
                self,
            ),
        )

    def __rrshift__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) in ArithmeticExpression.CONST_TYPES:
                return ArithmeticExpression(
                    self.op,
                    (
                        other >> self.operands[0],
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        other >> self.operands[1],
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                other,
                self,
            ),
        )

    def __rshift__(self, other):
        if type(other) in ArithmeticExpression.CONST_TYPES:
            other = self._unpack_const(other)
            if type(self.operands[0]) in ArithmeticExpression.CONST_TYPES:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0] >> other,
                        self.operands[1],
                    ),
                )
            elif type(self.operands[1]) is int:
                return ArithmeticExpression(
                    self.op,
                    (
                        self.operands[0],
                        self.operands[1] >> other,
                    ),
                )
        return ArithmeticExpression(
            self.op,
            (
                self,
                other,
            ),
        )

    @staticmethod
    def _unpack_const(expr):
        if type(expr) is int:
            return expr
        elif type(expr) is ailment.expression.Const:
            return expr.value
        raise NotImplementedError("Unsupported const expression type %s." % type(expr))

    @staticmethod
    def try_unpack_const(expr):
        try:
            return ArithmeticExpression._unpack_const(expr)
        except NotImplementedError:
            return expr


class RegisterOffset:
    __slots__ = (
        "_bits",
        "reg",
        "offset",
    )

    def __init__(self, bits, reg, offset):
        self._bits = bits
        self.reg = reg
        self.offset = offset

    @property
    def bits(self):
        return self._bits

    @property
    def symbolic(self):
        return type(self.offset) is not int

    def __repr__(self):
        if type(self.offset) is int:
            offset_str = "" if self.offset == 0 else "%+x" % self.offset
        else:
            offset_str = str(self.offset)
        return f"{self.reg}{offset_str}"

    def __add__(self, other):
        if not self.symbolic and type(other) is int:
            # Keep things in concrete
            return RegisterOffset(self._bits, self.reg, self._to_signed(self.offset + other))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, self.offset + other)
            else:
                # Convert to symbolic
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.Add,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(self.offset - other))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, self.offset - other)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.Sub,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __rsub__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(other - self.offset))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, other - self.offset)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.Sub,
                        (
                            other,
                            self.offset,
                        ),
                    ),
                )

    def __mul__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(self.offset * other))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, self.offset * other)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.Mul,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __rmul__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(other * self.offset))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, self.offset * other)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.Mul,
                        (
                            other,
                            self.offset,
                        ),
                    ),
                )

    def __and__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(self.offset & other))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, self.offset & other)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.And,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __rand__(self, other):
        return self.__and__(other)

    def __or__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(self.offset | other))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, self.offset | other)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.Or,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __ror__(self, other):
        return self.__or__(other)

    def __xor__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(self.offset | other))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, self.offset ^ other)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.Xor,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __rxor__(self, other):
        return self.__xor__(other)

    def __floordiv__(self, other):
        # this should never happen. returning self is obviously incorrect.
        return self

    def __rshift__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(self.offset >> other))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, self.offset >> other)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.RShift,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __rrshift__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(other >> self.offset))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, other >> self.offset)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.RShift,
                        (
                            other,
                            self.offset,
                        ),
                    ),
                )

    def __lshift__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(self.offset << other))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, self.offset << other)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.LShift,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __rlshift__(self, other):
        if not self.symbolic and type(other) is int:
            return RegisterOffset(self._bits, self.reg, self._to_signed(other << self.offset))
        else:
            if self.symbolic:
                return RegisterOffset(self._bits, self.reg, other << self.offset)
            else:
                return RegisterOffset(
                    self._bits,
                    self.reg,
                    ArithmeticExpression(
                        ArithmeticExpression.LShift,
                        (
                            other,
                            self.offset,
                        ),
                    ),
                )

    def __neg__(self):
        if not self.symbolic:
            return RegisterOffset(self._bits, self.reg, self._to_signed(-self.offset))
        else:
            return RegisterOffset(self._bits, self.reg, -self.offset)

    def __invert__(self):
        if not self.symbolic:
            return RegisterOffset(self._bits, self.reg, self._to_signed(~self.offset))
        else:
            return RegisterOffset(self._bits, self.reg, ~self.offset)

    def _to_signed(self, n):
        if n >= 2 ** (self._bits - 1):
            return n - 2**self._bits
        return n


class SpOffset(RegisterOffset):
    __slots__ = ("is_base",)

    def __init__(self, bits, offset, is_base=False):
        super().__init__(bits, "sp", offset)
        self.is_base = is_base

    def __repr__(self):
        if type(self.offset) is int:
            offset_str = "" if self.offset == 0 else "%+#x" % self.offset
        else:
            offset_str = str(self.offset)
        return "{}{}".format("BP" if self.is_base else "SP", offset_str)

    def __add__(self, other):
        other = ArithmeticExpression.try_unpack_const(other)
        if not self.symbolic and type(other) is int:
            return SpOffset(self._bits, self._to_signed(self.offset + other))
        else:
            if self.symbolic:
                return SpOffset(self._bits, self.offset + other)
            else:
                return SpOffset(
                    self._bits,
                    ArithmeticExpression(
                        ArithmeticExpression.Add,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            return self.offset - other.offset

        other = ArithmeticExpression.try_unpack_const(other)
        if not self.symbolic and type(other) is int:
            return SpOffset(self._bits, self._to_signed(self.offset - other))
        else:
            if self.symbolic:
                return SpOffset(self._bits, self.offset - other)
            else:
                return SpOffset(
                    self._bits,
                    ArithmeticExpression(
                        ArithmeticExpression.Sub,
                        (
                            self.offset,
                            other,
                        ),
                    ),
                )

    def __and__(self, other):
        other = ArithmeticExpression.try_unpack_const(other)
        if is_alignment_mask(other):
            # stack pointer alignment. ignore it.
            return SpOffset(self._bits, self.offset)
        else:
            return SpOffset(
                self._bits,
                ArithmeticExpression(
                    ArithmeticExpression.And,
                    (
                        self,
                        other,
                    ),
                ),
            )

    def __eq__(self, other):
        return (
            type(other) is SpOffset
            and self._bits == other.bits
            and self.reg == other.reg
            and self.offset == other.offset
            and self.is_base is other.is_base
        )

    def __hash__(self):
        return hash((self._bits, self.reg, self.offset, self.is_base))

    def __lt__(self, other):
        if type(other) is not SpOffset or self.reg != other.reg:
            return NotImplemented
        return self.offset < other.offset

    def __gt__(self, other):
        if type(other) is not SpOffset or self.reg != other.reg:
            return NotImplemented
        return self.offset > other.offset



import sys


def is_pyinstaller() -> bool:
    """
    Detect if we are currently running as a PyInstaller-packaged program.

    :return:    True if we are running as a PyInstaller-packaged program. False if we are running in Python directly
                (e.g., development mode).
    """
    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")



from typing import TYPE_CHECKING
import logging

import claripy
import pyvex

from ....utils.constants import DEFAULT_STATEMENT
from ....code_location import CodeLocation
from ....blade import Blade
from ...propagator import vex_vars
from .resolver import IndirectJumpResolver
from .propagator_utils import PropagatorLoadCallback

if TYPE_CHECKING:
    from angr import Block

_l = logging.getLogger(name=__name__)


def exists_in_replacements(replacements, block_loc, tmp_var):
    exists = False
    for rep in replacements:
        if rep == block_loc:
            exists = True
            break

    if not exists:
        return False

    exists = False
    for var in replacements[block_loc]:
        if var == tmp_var:
            exists = True
            break

    return exists


class ConstantResolver(IndirectJumpResolver):
    """
    Resolve an indirect jump by running a constant propagation on the entire function and check if the indirect jump can
    be resolved to a constant value. This resolver must be run after all other more specific resolvers.
    """

    def __init__(self, project):
        super().__init__(project, timeless=False)

    def filter(self, cfg, addr, func_addr, block, jumpkind):
        # we support both an indirect call and jump since the value can be resolved
        if jumpkind in {"Ijk_Boring", "Ijk_Call"}:
            return True

        return False

    def resolve(  # pylint:disable=unused-argument
        self, cfg, addr: int, func_addr: int, block: "Block", jumpkind: str, func_graph_complete: bool = True, **kwargs
    ):
        """
        This function does the actual resolve. Our process is easy:
        Propagate all values inside the function specified, then extract
        the tmp_var used for the indirect jump from the basic block.
        Use the tmp var to locate the constant value stored in the replacements.
        If not present, returns False tuple.

        :param cfg:         CFG with specified function
        :param addr:        Address of indirect jump
        :param func_addr:   Address of function of indirect jump
        :param block:       Block of indirect jump (Block object)
        :param jumpkind:    VEX jumpkind (Ijk_Boring or Ijk_Call)
        :return:            Bool tuple with replacement address
        """
        vex_block = block.vex
        if isinstance(vex_block.next, pyvex.expr.RdTmp):
            # what does the jump rely on? slice it back and see
            b = Blade(
                cfg.graph,
                addr,
                -1,
                cfg=cfg,
                project=self.project,
                ignore_bp=False,
                ignore_sp=False,
                max_level=3,
                stop_at_calls=True,
                cross_insn_opt=True,
            )
            stmt_loc = addr, DEFAULT_STATEMENT
            preds = list(b.slice.predecessors(stmt_loc))
            while preds:
                if len(preds) == 1:
                    # skip all IMarks
                    pred_addr, stmt_idx = preds[0]
                    if stmt_idx != DEFAULT_STATEMENT:
                        block = self.project.factory.block(pred_addr, cross_insn_opt=True).vex
                        if isinstance(block.statements[stmt_idx], pyvex.IRStmt.IMark):
                            preds = list(b.slice.predecessors(preds[0]))
                            continue

                for pred_addr, stmt_idx in preds:
                    block = self.project.factory.block(pred_addr, cross_insn_opt=True).vex
                    if stmt_idx != DEFAULT_STATEMENT:
                        stmt = block.statements[stmt_idx]
                        if isinstance(stmt, pyvex.IRStmt.WrTmp) and isinstance(stmt.data, pyvex.IRExpr.Load):
                            # loading from memory - unsupported
                            return False, []
                break

            func = cfg.functions[func_addr]
            _l.debug("ConstantResolver: Propagating for %r at %#x.", func, addr)
            prop = self.project.analyses.Propagator(
                func=func,
                only_consts=True,
                do_binops=True,
                vex_cross_insn_opt=False,
                completed_funcs=cfg._completed_functions,
                load_callback=PropagatorLoadCallback(self.project).propagator_load_callback,
                cache_results=True,
                key_prefix="cfg_intermediate",
            )

            replacements = prop.replacements
            if replacements:
                block_loc = CodeLocation(block.addr, None)
                tmp_var = vex_vars.VEXTmp(vex_block.next.tmp)

                if exists_in_replacements(replacements, block_loc, tmp_var):
                    resolved_tmp = replacements[block_loc][tmp_var]

                    if (
                        isinstance(resolved_tmp, claripy.ast.Base)
                        and resolved_tmp.op == "BVV"
                        and self._is_target_valid(cfg, resolved_tmp.args[0])
                    ):
                        return True, [resolved_tmp.args[0]]

        return False, []



import logging

from ..func import Func, TestData
from ....errors import SimMemoryError


l = logging.getLogger(name=__name__)


class realloc(Func):
    def __init__(self):
        super().__init__()  # pylint disable=useless-super-delegation

    def num_args(self):
        return 2

    def args(self):  # pylint disable=no-self-use
        return ["ptr", "size"]

    def get_name(self):
        return "realloc"

    def gen_input_output_pair(self):
        return None

    def pre_test(self, func, runner):
        # we should not get a nonzero output from the function with a value this large
        num = 0xFFFF0000
        test_input = [0, num]
        test_output = [None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        state = runner.get_out_state(func, test)
        if state is not None and state.solver.eval(state.regs.eax) != 0:
            return False

        # we should be able to get different outputs if we call malloc multiple times
        num = 0x80
        test_input = [0, num]
        test_output = [None, None]
        return_val = None

        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        returned_locs = []
        state = runner.get_out_state(func, test)
        if state is None:
            return False
        returned_locs.append(state.solver.eval(state.regs.eax))

        for i in range(10):  # pylint disable=unused-variable
            state = runner.get_out_state(func, test, initial_state=state)
            if state is None:
                return False
            returned_locs.append(state.solver.eval(state.regs.eax))

        # if we got the same value 2x it didnt work
        if len(set(returned_locs)) != len(returned_locs):
            return False

        # if we got 0 it didn't work
        if any(a == 0 for a in returned_locs):
            return False

        # if they are all multiples of 0x1000 it seems to be always calling allocate
        if all(a % 0x1000 == returned_locs[0] % 0x1000 for a in returned_locs):
            return False

        # they all should be writable/readable
        try:
            if any(state.solver.eval(state.memory.permissions(a)) & 3 != 3 for a in returned_locs):
                return False
        except SimMemoryError:
            return False

        # we should get different values if we try with a different size
        num = 0x320
        test_input = [0, num]
        test_output = [None, None]
        return_val = None
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        returned_locs2 = []
        state = runner.get_out_state(func, test)
        if state is None:
            return False
        returned_locs2.append(state.solver.eval(state.regs.eax))

        for i in range(10):
            state = runner.get_out_state(func, test, initial_state=state)
            if state is None:
                return False
            returned_locs2.append(state.solver.eval(state.regs.eax))

        if returned_locs == returned_locs2:
            return False

        return True



# pylint:disable=no-self-use,unused-argument
from typing import Set, Dict

import pyvex

from angr.engines.light import SimEngineLightVEXMixin


class VEXIRSBScanner(SimEngineLightVEXMixin):
    """
    Scan the VEX IRSB to determine if any argument-passing registers should be narrowed by detecting cases of loading
    the whole register and immediately narrowing the register before writing to the tmp.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # the following variables are for narrowing argument-passing register on 64-bit architectures. they are
        # initialized before processing each block.
        self.tmps_with_64bit_regs: Set[int] = set()  # tmps that store 64-bit register values
        self.tmps_converted_to_32bit: Set[int] = set()  # tmps that store the 64-to-32-bit converted values
        self.tmps_assignment_stmtidx: Dict[int, int] = {}  # statement IDs for the assignment of each tmp
        self.stmts_to_lower: Set[int] = set()

        # the following variables are for recognizing redundant argument register reads in gcc(?) -O0 binaries.
        # e.g.,
        #    mov rdi, r9
        #    mov rdi, rax
        #
        # we will not create a variable for the register read in the first instruction (read from r9) in this case.
        self.tmp_with_reg_as_value: Dict[int, int] = {}
        self.reg_with_reg_as_value: Dict[int, int] = {}
        self.reg_read_stmt_id: Dict[int, int] = {}
        self.reg_read_stmts_to_ignore: Set[int] = set()

    def _process_Stmt(self, whitelist=None):
        self.tmps_with_64bit_regs = set()
        self.tmps_assignment_stmtidx = {}
        self.tmps_converted_to_32bit = set()

        super()._process_Stmt(whitelist=whitelist)

        self.stmts_to_lower = {self.tmps_assignment_stmtidx[i] for i in self.tmps_converted_to_32bit}

    def _handle_Put(self, stmt):
        if isinstance(stmt.data, pyvex.IRExpr.RdTmp) and stmt.data.tmp in self.tmp_with_reg_as_value:
            if (
                stmt.offset in self.reg_with_reg_as_value
                and self.reg_with_reg_as_value[stmt.offset] != self.tmp_with_reg_as_value[stmt.data.tmp]
            ):
                # we are overwriting an existing register with a value from another register, before this register is
                # ever used...
                # in this case, we should ignore the previous register read
                old_reg_offset = self.reg_with_reg_as_value[stmt.offset]
                self.reg_read_stmts_to_ignore.add(self.reg_read_stmt_id[old_reg_offset])
            self.reg_with_reg_as_value[stmt.offset] = self.tmp_with_reg_as_value[stmt.data.tmp]

    def _handle_Load(self, expr):
        pass

    def _handle_Store(self, stmt):
        pass

    def _handle_LoadG(self, stmt):
        pass

    def _handle_LLSC(self, stmt: pyvex.IRStmt.LLSC):
        pass

    def _handle_StoreG(self, stmt):
        pass

    def _handle_WrTmp(self, stmt):
        if isinstance(stmt.data, pyvex.IRExpr.Get) and stmt.data.result_size(self.tyenv) == 64:
            self.tmps_with_64bit_regs.add(stmt.tmp)
        self.tmps_assignment_stmtidx[stmt.tmp] = self.stmt_idx

        if isinstance(stmt.data, pyvex.IRExpr.Get) and stmt.data.result_size(self.tyenv) == 64:
            self.tmp_with_reg_as_value[stmt.tmp] = stmt.data.offset

        super()._handle_WrTmp(stmt)

    def _handle_Get(self, expr):
        self.reg_read_stmt_id[expr.offset] = self.stmt_idx
        if expr.offset in self.reg_with_reg_as_value:
            del self.reg_with_reg_as_value[expr.offset]

    def _handle_RdTmp(self, expr):
        if expr.tmp in self.tmps_converted_to_32bit:
            self.tmps_converted_to_32bit.remove(expr.tmp)

    def _handle_Conversion(self, expr: pyvex.IRExpr.Unop):
        if expr.op == "Iop_64to32" and isinstance(expr.args[0], pyvex.IRExpr.RdTmp):
            # special handling for t11 = GET:I64(rdi); t4 = 64to32(t11) style of code in x86-64 (and other 64-bit
            # architectures as well)
            tmp_src = expr.args[0].tmp
            if tmp_src in self.tmps_with_64bit_regs:
                self.tmps_converted_to_32bit.add(tmp_src)

    def _handle_16HLto32(self, expr):
        pass

    def _handle_Cmp_v(self, expr, _vector_size, _vector_count):
        pass

    _handle_CmpEQ_v = _handle_Cmp_v
    _handle_CmpNE_v = _handle_Cmp_v
    _handle_CmpLE_v = _handle_Cmp_v
    _handle_CmpLT_v = _handle_Cmp_v
    _handle_CmpGE_v = _handle_Cmp_v
    _handle_CmpGT_v = _handle_Cmp_v

    def _handle_ExpCmpNE64(self, expr):
        pass

    def _handle_CCall(self, expr):
        pass

    def _handle_function(self, func_addr):
        pass



from . import SimHeapBase


class SimHeapLibc(SimHeapBase):
    """
    A class of heap that implements the major libc heap management functions.
    """

    def malloc(self, sim_size):
        """
        A somewhat faithful implementation of libc `malloc`.

        :param sim_size: the amount of memory (in bytes) to be allocated
        :returns:        the address of the allocation, or a NULL pointer if the allocation failed
        """
        raise NotImplementedError(f"{self.malloc.__func__.__name__} not implemented for {self.__class__.__name__}")

    def free(self, ptr):  # pylint:disable=unused-argument
        """
        A somewhat faithful implementation of libc `free`.

        :param ptr: the location in memory to be freed
        """
        raise NotImplementedError(f"{self.free.__func__.__name__} not implemented for {self.__class__.__name__}")

    def calloc(self, sim_nmemb, sim_size):
        """
        A somewhat faithful implementation of libc `calloc`.

        :param     sim_nmemb: the number of elements to allocated
        :param     sim_size: the size of each element (in bytes)
        :returns:  the address of the allocation, or a NULL pointer if the allocation failed
        """
        raise NotImplementedError(f"{self.calloc.__func__.__name__} not implemented for {self.__class__.__name__}")

    def realloc(self, ptr, size):
        """
        A somewhat faithful implementation of libc `realloc`.

        :param ptr:  the location in memory to be reallocated
        :param size: the new size desired for the allocation
        :returns:    the address of the allocation, or a NULL pointer if the allocation was freed or if no new
                     allocation was made
        """
        raise NotImplementedError(f"{self.realloc.__func__.__name__} not implemented for {self.__class__.__name__}")



import angr


class arch_prctl(angr.SimProcedure):
    """
    Sets the architecture specific thread state based on the subfunction selected
    using the 'code' parameter. This syscall is only present on x86_64 linux
    """

    def run(self, code, addr):  # pylint: disable=arguments-differ
        if self.state.solver.symbolic(code):
            raise angr.errors.SimValueError("Code value passed to arch_prctl must be concrete.")

        code = self.state.solver.eval(code)

        # ARCH_SET_GS
        if code == 0x1001:
            self.state.regs.gs = addr
        # ARCH_SET_FS
        elif code == 0x1002:
            self.state.regs.fs = addr
        # ARCH_GET_FS
        elif code == 0x1003:
            fs = self.state.regs.fs
            self.state.memory.store(addr, fs)
        # ARCH_GET_GS
        elif code == 0x1004:
            gs = self.state.regs.gs
            self.state.memory.store(addr, gs)
        else:
            # EINVAL is returned if code is not a valid subcommand.
            return 22
        return 0



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("msvfw32.dll")
prototypes = \
    {
        #
        'VideoForWindowsVersion': SimTypeFunction([], SimTypeInt(signed=False, label="UInt32")),
        #
        'ICInfo': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwVersionICM": SimTypeInt(signed=False, label="UInt32"), "szName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 16), "szDescription": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 128), "szDriver": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 128)}, name="ICINFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["fccType", "fccHandler", "lpicinfo"]),
        #
        'ICInstall': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["fccType", "fccHandler", "lParam", "szDesc", "wFlags"]),
        #
        'ICRemove': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["fccType", "fccHandler", "wFlags"]),
        #
        'ICGetInfo': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwVersionICM": SimTypeInt(signed=False, label="UInt32"), "szName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 16), "szDescription": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 128), "szDriver": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 128)}, name="ICINFO", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hic", "picinfo", "cb"]),
        #
        'ICOpen': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["fccType", "fccHandler", "wMode"]),
        #
        'ICOpenFunction': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["fccType", "fccHandler", "wMode", "lpfnHandler"]),
        #
        'ICClose': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hic"]),
        #
        'ICSendMessage': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hic", "msg", "dw1", "dw2"]),
        #
        'ICCompress': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hic", "dwFlags", "lpbiOutput", "lpData", "lpbiInput", "lpBits", "lpckid", "lpdwFlags", "lFrameNum", "dwFrameSize", "dwQuality", "lpbiPrev", "lpPrev"]),
        #
        'ICDecompress': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hic", "dwFlags", "lpbiFormat", "lpData", "lpbi", "lpBits"]),
        #
        'ICDrawBegin': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hic", "dwFlags", "hpal", "hwnd", "hdc", "xDst", "yDst", "dxDst", "dyDst", "lpbi", "xSrc", "ySrc", "dxSrc", "dySrc", "dwRate", "dwScale"]),
        #
        'ICDraw': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hic", "dwFlags", "lpFormat", "lpData", "cbData", "lTime"]),
        #
        'ICLocate': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypeShort(signed=False, label="UInt16")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["fccType", "fccHandler", "lpbiIn", "lpbiOut", "wFlags"]),
        #
        'ICGetDisplayFormat': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hic", "lpbiIn", "lpbiOut", "BitDepth", "dx", "dy"]),
        #
        'ICImageCompress': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hic", "uiFlags", "lpbiIn", "lpBits", "lpbiOut", "lQuality", "plSize"]),
        #
        'ICImageDecompress': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hic", "uiFlags", "lpbiIn", "lpBits", "lpbiOut"]),
        #
        'ICCompressorChoose': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=True, label="Int32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "hic": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "lpbiIn": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpbiOut": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpBitsOut": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lpBitsPrev": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lFrame": SimTypeInt(signed=True, label="Int32"), "lKey": SimTypeInt(signed=True, label="Int32"), "lDataRate": SimTypeInt(signed=True, label="Int32"), "lQ": SimTypeInt(signed=True, label="Int32"), "lKeyCount": SimTypeInt(signed=True, label="Int32"), "lpState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbState": SimTypeInt(signed=True, label="Int32")}, name="COMPVARS", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "uiFlags", "pvIn", "lpData", "pc", "lpszTitle"]),
        #
        'ICSeqCompressFrameStart': SimTypeFunction([SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=True, label="Int32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "hic": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "lpbiIn": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpbiOut": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpBitsOut": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lpBitsPrev": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lFrame": SimTypeInt(signed=True, label="Int32"), "lKey": SimTypeInt(signed=True, label="Int32"), "lDataRate": SimTypeInt(signed=True, label="Int32"), "lQ": SimTypeInt(signed=True, label="Int32"), "lKeyCount": SimTypeInt(signed=True, label="Int32"), "lpState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbState": SimTypeInt(signed=True, label="Int32")}, name="COMPVARS", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pc", "lpbiIn"]),
        #
        'ICSeqCompressFrameEnd': SimTypeFunction([SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=True, label="Int32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "hic": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "lpbiIn": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpbiOut": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpBitsOut": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lpBitsPrev": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lFrame": SimTypeInt(signed=True, label="Int32"), "lKey": SimTypeInt(signed=True, label="Int32"), "lDataRate": SimTypeInt(signed=True, label="Int32"), "lQ": SimTypeInt(signed=True, label="Int32"), "lKeyCount": SimTypeInt(signed=True, label="Int32"), "lpState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbState": SimTypeInt(signed=True, label="Int32")}, name="COMPVARS", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pc"]),
        #
        'ICSeqCompressFrame': SimTypeFunction([SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=True, label="Int32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "hic": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "lpbiIn": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpbiOut": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpBitsOut": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lpBitsPrev": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lFrame": SimTypeInt(signed=True, label="Int32"), "lKey": SimTypeInt(signed=True, label="Int32"), "lDataRate": SimTypeInt(signed=True, label="Int32"), "lQ": SimTypeInt(signed=True, label="Int32"), "lKeyCount": SimTypeInt(signed=True, label="Int32"), "lpState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbState": SimTypeInt(signed=True, label="Int32")}, name="COMPVARS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["pc", "uiFlags", "lpBits", "pfKey", "plSize"]),
        #
        'ICCompressorFree': SimTypeFunction([SimTypePointer(SimStruct({"cbSize": SimTypeInt(signed=True, label="Int32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "hic": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "lpbiIn": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpbiOut": SimTypePointer(SimStruct({"bmiHeader": SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), "bmiColors": SimTypePointer(SimStruct({"rgbBlue": SimTypeChar(label="Byte"), "rgbGreen": SimTypeChar(label="Byte"), "rgbRed": SimTypeChar(label="Byte"), "rgbReserved": SimTypeChar(label="Byte")}, name="RGBQUAD", pack=False, align=None), offset=0)}, name="BITMAPINFO", pack=False, align=None), offset=0), "lpBitsOut": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lpBitsPrev": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "lFrame": SimTypeInt(signed=True, label="Int32"), "lKey": SimTypeInt(signed=True, label="Int32"), "lDataRate": SimTypeInt(signed=True, label="Int32"), "lQ": SimTypeInt(signed=True, label="Int32"), "lKeyCount": SimTypeInt(signed=True, label="Int32"), "lpState": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbState": SimTypeInt(signed=True, label="Int32")}, name="COMPVARS", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pc"]),
        #
        'DrawDibOpen': SimTypeFunction([], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)),
        #
        'DrawDibClose': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdd"]),
        #
        'DrawDibGetBuffer': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["hdd", "lpbi", "dwSize", "dwFlags"]),
        #
        'DrawDibGetPalette': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hdd"]),
        #
        'DrawDibSetPalette': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdd", "hpal"]),
        #
        'DrawDibChangePalette': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"peRed": SimTypeChar(label="Byte"), "peGreen": SimTypeChar(label="Byte"), "peBlue": SimTypeChar(label="Byte"), "peFlags": SimTypeChar(label="Byte")}, name="PALETTEENTRY", pack=False, align=None), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdd", "iStart", "iLen", "lppe"]),
        #
        'DrawDibRealize': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hdd", "hdc", "fBackground"]),
        #
        'DrawDibStart': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hdd", "rate"]),
        #
        'DrawDibStop': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdd"]),
        #
        'DrawDibBegin': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hdd", "hdc", "dxDst", "dyDst", "lpbi", "dxSrc", "dySrc", "wFlags"]),
        #
        'DrawDibDraw': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hdd", "hdc", "xDst", "yDst", "dxDst", "dyDst", "lpbi", "lpBits", "xSrc", "ySrc", "dxSrc", "dySrc", "wFlags"]),
        #
        'DrawDibEnd': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdd"]),
        #
        'DrawDibTime': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"timeCount": SimTypeInt(signed=True, label="Int32"), "timeDraw": SimTypeInt(signed=True, label="Int32"), "timeDecompress": SimTypeInt(signed=True, label="Int32"), "timeDither": SimTypeInt(signed=True, label="Int32"), "timeStretch": SimTypeInt(signed=True, label="Int32"), "timeBlt": SimTypeInt(signed=True, label="Int32"), "timeSetDIBits": SimTypeInt(signed=True, label="Int32")}, name="DRAWDIBTIME", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdd", "lpddtime"]),
        #
        'DrawDibProfileDisplay': SimTypeFunction([SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["lpbi"]),
        #
        'MCIWndCreateA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hwndParent", "hInstance", "dwStyle", "szFile"]),
        #
        'MCIWndCreateW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hwndParent", "hInstance", "dwStyle", "szFile"]),
        #
        'MCIWndRegisterClass': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'GetOpenFileNamePreviewA': SimTypeFunction([SimTypePointer(SimStruct({"lStructSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hInstance": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpstrFilter": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lpstrCustomFilter": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nMaxCustFilter": SimTypeInt(signed=False, label="UInt32"), "nFilterIndex": SimTypeInt(signed=False, label="UInt32"), "lpstrFile": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nMaxFile": SimTypeInt(signed=False, label="UInt32"), "lpstrFileTitle": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nMaxFileTitle": SimTypeInt(signed=False, label="UInt32"), "lpstrInitialDir": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lpstrTitle": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Flags": SimTypeInt(signed=False, label="OPEN_FILENAME_FLAGS"), "nFileOffset": SimTypeShort(signed=False, label="UInt16"), "nFileExtension": SimTypeShort(signed=False, label="UInt16"), "lpstrDefExt": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lCustData": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpfnHook": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), arg_names=["param0", "param1", "param2", "param3"]), offset=0), "lpTemplateName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pvReserved": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "FlagsEx": SimTypeInt(signed=False, label="OPEN_FILENAME_FLAGS_EX")}, name="OPENFILENAMEA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpofn"]),
        #
        'GetSaveFileNamePreviewA': SimTypeFunction([SimTypePointer(SimStruct({"lStructSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hInstance": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpstrFilter": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lpstrCustomFilter": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nMaxCustFilter": SimTypeInt(signed=False, label="UInt32"), "nFilterIndex": SimTypeInt(signed=False, label="UInt32"), "lpstrFile": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nMaxFile": SimTypeInt(signed=False, label="UInt32"), "lpstrFileTitle": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nMaxFileTitle": SimTypeInt(signed=False, label="UInt32"), "lpstrInitialDir": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lpstrTitle": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Flags": SimTypeInt(signed=False, label="OPEN_FILENAME_FLAGS"), "nFileOffset": SimTypeShort(signed=False, label="UInt16"), "nFileExtension": SimTypeShort(signed=False, label="UInt16"), "lpstrDefExt": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "lCustData": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpfnHook": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), arg_names=["param0", "param1", "param2", "param3"]), offset=0), "lpTemplateName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pvReserved": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "FlagsEx": SimTypeInt(signed=False, label="OPEN_FILENAME_FLAGS_EX")}, name="OPENFILENAMEA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpofn"]),
        #
        'GetOpenFileNamePreviewW': SimTypeFunction([SimTypePointer(SimStruct({"lStructSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hInstance": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpstrFilter": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpstrCustomFilter": SimTypePointer(SimTypeChar(label="Char"), offset=0), "nMaxCustFilter": SimTypeInt(signed=False, label="UInt32"), "nFilterIndex": SimTypeInt(signed=False, label="UInt32"), "lpstrFile": SimTypePointer(SimTypeChar(label="Char"), offset=0), "nMaxFile": SimTypeInt(signed=False, label="UInt32"), "lpstrFileTitle": SimTypePointer(SimTypeChar(label="Char"), offset=0), "nMaxFileTitle": SimTypeInt(signed=False, label="UInt32"), "lpstrInitialDir": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpstrTitle": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Flags": SimTypeInt(signed=False, label="OPEN_FILENAME_FLAGS"), "nFileOffset": SimTypeShort(signed=False, label="UInt16"), "nFileExtension": SimTypeShort(signed=False, label="UInt16"), "lpstrDefExt": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lCustData": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpfnHook": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), arg_names=["param0", "param1", "param2", "param3"]), offset=0), "lpTemplateName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pvReserved": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "FlagsEx": SimTypeInt(signed=False, label="OPEN_FILENAME_FLAGS_EX")}, name="OPENFILENAMEW", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpofn"]),
        #
        'GetSaveFileNamePreviewW': SimTypeFunction([SimTypePointer(SimStruct({"lStructSize": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "hInstance": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpstrFilter": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpstrCustomFilter": SimTypePointer(SimTypeChar(label="Char"), offset=0), "nMaxCustFilter": SimTypeInt(signed=False, label="UInt32"), "nFilterIndex": SimTypeInt(signed=False, label="UInt32"), "lpstrFile": SimTypePointer(SimTypeChar(label="Char"), offset=0), "nMaxFile": SimTypeInt(signed=False, label="UInt32"), "lpstrFileTitle": SimTypePointer(SimTypeChar(label="Char"), offset=0), "nMaxFileTitle": SimTypeInt(signed=False, label="UInt32"), "lpstrInitialDir": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lpstrTitle": SimTypePointer(SimTypeChar(label="Char"), offset=0), "Flags": SimTypeInt(signed=False, label="OPEN_FILENAME_FLAGS"), "nFileOffset": SimTypeShort(signed=False, label="UInt16"), "nFileExtension": SimTypeShort(signed=False, label="UInt16"), "lpstrDefExt": SimTypePointer(SimTypeChar(label="Char"), offset=0), "lCustData": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpfnHook": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), arg_names=["param0", "param1", "param2", "param3"]), offset=0), "lpTemplateName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pvReserved": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "FlagsEx": SimTypeInt(signed=False, label="OPEN_FILENAME_FLAGS_EX")}, name="OPENFILENAMEW", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpofn"]),
    }

lib.set_prototypes(prototypes)



import angr

from cle.backends.externs.simdata.io_file import io_file_data_for_arch


class feof(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, file_ptr):
        # TODO handle errors
        fd_offset = io_file_data_for_arch(self.state.arch)["fd"]
        fileno = self.state.mem[file_ptr + fd_offset :].int.concrete
        simfd = self.state.posix.get_fd(fileno)
        if simfd is None:
            return None
        return self.state.solver.If(simfd.eof(), self.state.solver.BVV(1, self.arch.sizeof["int"]), 0)


feof_unlocked = feof



def slice_callgraph(callgraph, cfg_slice_to_sink):
    """
    Slice a callgraph, keeping only the nodes present in the <CFGSliceToSink> representation, and th transitions for
    which a path exists.

    *Note* that this function mutates the graph passed as an argument.

    :param networkx.MultiDiGraph callgraph:
        The callgraph to update.
    :param CFGSliceToSink cfg_slice_to_sink:
        The representation of the slice, containing the data to update the callgraph from.
    """

    edges_to_remove = list(filter(lambda e: not cfg_slice_to_sink.path_between(*e), callgraph.edges()))

    nodes_to_remove = list(filter(lambda node: node not in cfg_slice_to_sink.nodes, callgraph.nodes()))

    callgraph.remove_edges_from(edges_to_remove)
    callgraph.remove_nodes_from(nodes_to_remove)

    return callgraph


def slice_cfg_graph(graph, cfg_slice_to_sink):
    """
    Slice a CFG graph, keeping only the transitions and nodes present in the <CFGSliceToSink> representation.

    *Note* that this function mutates the graph passed as an argument.

    :param networkx.DiGraph graph: The graph to slice.
    :param CFGSliceToSink cfg_slice_to_sink:
        The representation of the slice, containing the data to update the CFG from.

    :return networkx.DiGraph: The sliced graph.
    """

    def _edge_in_slice_transitions(transitions, edge):
        if edge[0].addr not in transitions.keys():
            return False
        return edge[1].addr in cfg_slice_to_sink.transitions[edge[0].addr]

    edges_to_remove = list(
        filter(lambda edge: not _edge_in_slice_transitions(cfg_slice_to_sink.transitions, edge), graph.edges())
    )

    nodes_to_remove = list(filter(lambda node: node.addr not in cfg_slice_to_sink.nodes, graph.nodes()))

    graph.remove_edges_from(edges_to_remove)
    graph.remove_nodes_from(nodes_to_remove)

    return graph


def slice_function_graph(function_graph, cfg_slice_to_sink):
    """
    Slice a function graph, keeping only the nodes present in the <CFGSliceToSink> representation.

    Because the <CFGSliceToSink> is build from the CFG, and the function graph is *NOT* a subgraph of the CFG, edges of
    the function graph will no be present in the <CFGSliceToSink> transitions.
    However, we use the fact that if there is an edge between two nodes in the function graph, then there must exist
    a path between these two nodes in the slice; Proof idea:
    - The <CFGSliceToSink> is backward and recursively constructed;
    - If a node is in the slice, then all its predecessors will be (transitively);
    - If there is an edge between two nodes in the function graph, there is a path between them in the CFG;
    - So: The origin node is a transitive predecessor of the destination one, hence if destination is in the slice,
    then origin will be too.

    In consequence, in the end, removing the only nodes not present in the slice, and their related transitions gives
    us the expected result: a function graph representing (a higher view of) the flow in the slice.

    *Note* that this function mutates the graph passed as an argument.

    :param networkx.DiGraph graph: The graph to slice.
    :param CFGSliceToSink cfg_slice_to_sink:
        The representation of the slice, containing the data to update the CFG from.

    :return networkx.DiGraph: The sliced graph.
    """

    nodes_to_remove = list(filter(lambda node: node.addr not in cfg_slice_to_sink.nodes, function_graph.nodes()))

    function_graph.remove_nodes_from(nodes_to_remove)

    return function_graph



import itertools

import angr

rand_count = itertools.count()


class random(angr.SimProcedure):
    # pylint:disable=arguments-differ,missing-class-docstring

    def run(self, buf, count, rnd_bytes, concrete_data=None):
        if self.state.mode == "fastpath":
            # Special case for CFG
            if (
                not self.state.solver.symbolic(count)
                and not self.state.solver.symbolic(buf)
                and not self.state.solver.symbolic(rnd_bytes)
            ):
                if (
                    self.state.solver.is_true(rnd_bytes != 0) and self.state.cgc.addr_invalid(rnd_bytes)
                ) or self.state.cgc.addr_invalid(buf):
                    return self.state.cgc.EFAULT

                max_count = self.state.solver.eval_one(count)
                random_num = self.state.solver.Unconstrained(f"random_{next(rand_count)}", max_count * 8)
                self.state.memory.store(buf, random_num, size=count)
                if self.state.solver.is_true(rnd_bytes != 0):
                    self.state.memory.store(rnd_bytes, count, endness="Iend_LE")

            # We always return something in fastpath mode
            return self.state.solver.BVV(0, self.state.arch.bits)

        # return code
        r = self.state.solver.ite_cases(
            (
                (self.state.cgc.addr_invalid(buf), self.state.cgc.EFAULT),
                (self.state.solver.And(rnd_bytes != 0, self.state.cgc.addr_invalid(rnd_bytes)), self.state.cgc.EFAULT),
            ),
            self.state.solver.BVV(0, self.state.arch.bits),
        )

        if self.state.satisfiable(extra_constraints=[count != 0]):
            max_size = min(
                self.state.solver.max_int(count * 8),
                0x10000,
            )

            if concrete_data:
                value = self.state.solver.BVS(f"random_{next(rand_count)}", max_size)
                self.state.preconstrainer.preconstrain(concrete_data, value)
            else:
                value = self.state.solver.Unconstrained(
                    f"random_{next(rand_count)}", max_size, key=("syscall", "random")
                )

            self.state.memory.store(buf, value, size=count)

        self.state.memory.store(rnd_bytes, count, endness="Iend_LE", condition=rnd_bytes != 0)

        return r



# pylint:disable=no-member,global-statement
import os
import time
from functools import wraps
from collections import defaultdict

TIMING = os.environ.get("TIMING", "").lower() not in {"", "no", "0", "false"}
PRINT = os.environ.get("PRINT", "").lower() not in {"", "no", "0", "false"}
TIME_DISTRIBUTION = False

total_time = defaultdict(float)
time_distribution = defaultdict(list)
depth = 0


def timethis(func):
    @wraps(func)
    def timed_func(*args, **kwargs):
        if TIMING:

            def _t():
                return time.perf_counter_ns() / 1000000

            global depth
            depth += 1

            start = _t()
            r = func(*args, **kwargs)
            millisec = _t() - start
            sec = millisec / 1000
            if PRINT:
                indent = " " * ((depth - 1) * 2)
                if sec > 1.0:
                    print(f"[timing] {indent}{func.__name__} took {sec:f} seconds ({millisec:f} milliseconds).")
                else:
                    print(f"[timing] {indent}{func.__name__} took {millisec:f} milliseconds.")
            total_time[func] += millisec
            if TIME_DISTRIBUTION:
                time_distribution[func].append(millisec)
            depth -= 1
            return r
        else:
            return func(*args, **kwargs)

    return timed_func



import claripy
from archinfo.arch_arm import is_arm_arch
from . import MemoryMixin

stn_map = {"st%d" % n: n for n in range(8)}
tag_map = {"tag%d" % n: n for n in range(8)}


class NameResolutionMixin(MemoryMixin):
    """
    This mixin allows you to provide register names as load addresses, and will automatically translate this to an
    offset and size.
    """

    def _resolve_location_name(self, name, is_write=False):
        # Delayed load so SimMemory does not rely on SimEngines
        from ...engines.vex.claripy.ccall import _get_flags

        if self.category == "reg":
            if self.state.arch.name in ("X86", "AMD64"):
                if name in stn_map:
                    return (((stn_map[name] + self.load("ftop")) & 7) << 3) + self.state.arch.registers["fpu_regs"][
                        0
                    ], 8
                elif name in tag_map:
                    return ((tag_map[name] + self.load("ftop")) & 7) + self.state.arch.registers["fpu_tags"][0], 1
                elif name in ("flags", "eflags", "rflags"):
                    # we tweak the state to convert the vex condition registers into the flags register
                    if not is_write:  # this work doesn't need to be done if we're just gonna overwrite it
                        self.store("cc_dep1", _get_flags(self.state))  # constraints cannot be added by this
                    self.store("cc_op", 0)  # OP_COPY
                    return self.state.arch.registers["cc_dep1"]
            if is_arm_arch(self.state.arch):
                if name == "flags":
                    if not is_write:
                        self.store("cc_dep1", _get_flags(self.state))
                    self.store("cc_op", 0)
                    return self.state.arch.registers["cc_dep1"]

            if name == "sp" and "sp" not in self.state.arch.registers:
                sp_reg_name = self.state.arch.register_names[self.state.arch.sp_offset]
                return self.state.arch.registers[sp_reg_name]
            if name == "lr" and "lr" not in self.state.arch.registers:
                lr_reg_name = self.state.arch.register_names[self.state.arch.lr_offset]
                return self.state.arch.registers[lr_reg_name]

            return self.state.arch.registers[name]
        elif name[0] == "*":
            return self.state.registers.load(name[1:]), None
        else:
            raise SimMemoryError("Trying to address memory with a register name.")

    def store(self, addr, data, size=None, **kwargs):
        if isinstance(addr, str):
            named_addr, named_size = self._resolve_location_name(addr, is_write=True)
            if isinstance(data, claripy.ast.BV) and len(data) < named_size * self.state.arch.byte_width:
                data = data.zero_extend(named_size * self.state.arch.byte_width - len(data))
            return super().store(named_addr, data, size=named_size if size is None else size, **kwargs)
        else:
            return super().store(addr, data, size=size, **kwargs)

    def load(self, addr, size=None, **kwargs):
        if isinstance(addr, str):
            named_addr, named_size = self._resolve_location_name(addr, is_write=False)
            return super().load(named_addr, size=named_size if size is None else size, **kwargs)
        else:
            return super().load(addr, size=size, **kwargs)


from ...errors import SimMemoryError



# pylint:disable=arguments-differ
from typing import Tuple, Optional
import string

from archinfo import Endness

from ailment.expression import Const
from ailment.statement import Call, Store

from .base import PeepholeOptimizationStmtBase


ASCII_PRINTABLES = set(string.printable)
ASCII_DIGITS = set(string.digits)


class InlinedStrcpy(PeepholeOptimizationStmtBase):
    """
    Simplifies inlined string copying logic into calls to strcpy.
    """

    __slots__ = ()

    NAME = "Simplifying inlined strcpy"
    stmt_classes = (Store,)

    def optimize(self, stmt: Store, **kwargs):
        if isinstance(stmt.data, Const):
            r, s = self.is_integer_likely_a_string(stmt.data.value, stmt.data.size, stmt.endness)
            if r:
                # replace it with a call to strncpy
                str_id = self.kb.custom_strings.allocate(s.encode("ascii"))
                return Call(
                    stmt.idx,
                    "strncpy",
                    args=[
                        stmt.addr,
                        Const(None, None, str_id, stmt.addr.bits, custom_string=True),
                        Const(None, None, len(s), self.project.arch.bits),
                    ],
                    **stmt.tags,
                )

        return None

    @staticmethod
    def is_integer_likely_a_string(
        v: int, size: int, endness: Endness, min_length: int = 4
    ) -> Tuple[bool, Optional[str]]:
        # we need at least four bytes of printable characters

        chars = []
        if endness == Endness.LE:
            while v != 0:
                byt = v & 0xFF
                if chr(byt) not in ASCII_PRINTABLES:
                    return False, None
                chars.append(chr(byt))
                v >>= 8

        elif endness == Endness.BE:
            first_non_zero = False
            for _ in range(size):
                byt = v & 0xFF
                v >>= 8
                if byt == 0:
                    if first_non_zero:
                        return False, None
                    continue
                first_non_zero = True  # this is the first non-zero byte
                if chr(byt) not in ASCII_PRINTABLES:
                    return False, None
                chars.append(chr(byt))
            chars = chars[::-1]
        else:
            # unsupported endness
            return False, None

        if len(chars) >= min_length:
            if len(chars) <= 4 and all(ch in ASCII_DIGITS for ch in chars):
                return False, None
            return True, "".join(chars)
        return False, None



import angr
import tempfile


class tmpnam(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, tmp_file_path_addr):
        L_tmpnam = 20

        if self.state.solver.eval(tmp_file_path_addr) != 0:
            return tmp_file_path_addr

        tmp_file_path = tempfile.mktemp()
        malloc = angr.SIM_PROCEDURES["libc"]["malloc"]
        addr = self.inline_call(malloc, L_tmpnam).ret_expr
        self.state.memory.store(addr, tmp_file_path.encode() + b"\x00")

        return addr



import angr
from angr.sim_options import MEMORY_CHUNK_INDIVIDUAL_READS

import logging

l = logging.getLogger(name=__name__)


class strstr(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, haystack_addr, needle_addr, haystack_strlen=None, needle_strlen=None):
        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]
        strncmp = angr.SIM_PROCEDURES["libc"]["strncmp"]

        haystack_strlen = self.inline_call(strlen, haystack_addr) if haystack_strlen is None else haystack_strlen
        needle_strlen = self.inline_call(strlen, needle_addr) if needle_strlen is None else needle_strlen

        # naive approach
        haystack_maxlen = haystack_strlen.max_null_index
        needle_maxlen = needle_strlen.max_null_index

        l.debug("strstr with size %d haystack and size %d needle...", haystack_maxlen, needle_maxlen)

        if needle_maxlen == 0:
            l.debug("... zero-length needle.")
            return haystack_addr
        elif haystack_maxlen == 0:
            l.debug("... zero-length haystack.")
            return self.state.solver.BVV(0, self.state.arch.bits)

        if self.state.solver.symbolic(needle_strlen.ret_expr):
            cases = [[needle_strlen.ret_expr == 0, haystack_addr]]
            exclusions = [needle_strlen.ret_expr != 0]
            not_terminated_yet_constraints = []
            remaining_symbolic = self.state.libc.max_symbolic_strstr

            for i in range(haystack_maxlen):
                l.debug("... case %d (%d symbolic checks remaining)", i, remaining_symbolic)
                not_terminated_yet_constraints.append(self.state.memory.load(haystack_addr + i, 1) != 0)

                # big hack!
                cmp_res = self.inline_call(
                    strncmp,
                    haystack_addr + i,
                    needle_addr,
                    needle_strlen.ret_expr,
                    a_len=haystack_strlen,
                    b_len=needle_strlen,
                )

                c = self.state.solver.And(
                    *(
                        [self.state.solver.UGE(haystack_strlen.ret_expr, needle_strlen.ret_expr), cmp_res.ret_expr == 0]
                        + not_terminated_yet_constraints
                        + exclusions
                    )
                )

                exclusions.append(cmp_res.ret_expr != 0)

                if self.state.solver.symbolic(c):
                    remaining_symbolic -= 1

                cases.append([c, haystack_addr + i])
                haystack_strlen.ret_expr = haystack_strlen.ret_expr - 1

                if remaining_symbolic == 0:
                    l.debug("... exhausted remaining symbolic checks.")
                    break

            cases.append([self.state.solver.And(*exclusions), self.state.solver.BVV(0, self.state.arch.bits)])
            l.debug("... created %d cases", len(cases))
            r = self.state.solver.ite_cases(cases, 0)
            c = [self.state.solver.Or(*[c for c, _ in cases])]
        else:
            needle_length = self.state.solver.eval(needle_strlen.ret_expr)
            needle_str = self.state.memory.load(needle_addr, needle_length)

            chunk_size = None
            if MEMORY_CHUNK_INDIVIDUAL_READS in self.state.options:
                chunk_size = 1

            r, c, i = self.state.memory.find(
                haystack_addr,
                needle_str,
                haystack_strlen.max_null_index,
                max_symbolic_bytes=self.state.libc.max_symbolic_strstr,
                default=0,
                chunk_size=chunk_size,
            )

        self.state.add_constraints(*c)
        return r



import angr


class crazy_scanf(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, src, fmt, one, two, three):  # pylint:disable=unused-argument
        memcpy = angr.SIM_PROCEDURES["libc"]["memcpy"]

        self.inline_call(memcpy, one, src, 5)
        self.state.memory.store(one + 4, self.state.solver.BVV(0, 8))
        self.inline_call(memcpy, two, src + 6, 8192)
        self.state.memory.store(two + 8191, self.state.solver.BVV(0, 8))
        self.inline_call(memcpy, three, src + 6 + 8193, 12)
        self.state.memory.store(three + 11, self.state.solver.BVV(0, 8))

        return self.state.solver.BVV(3)



import logging

l = logging.getLogger("angr.engines.soot.expressions")


def translate_expr(expr, state):
    expr_name = expr.__class__.__name__.split(".")[-1]
    if expr_name.startswith("Soot"):
        expr_name = expr_name[4:]
    if expr_name.endswith("Expr"):
        expr_name = expr_name[:-4]
    expr_cls_name = "SimSootExpr_" + expr_name

    g = globals()
    if expr_cls_name in g:
        expr_cls = g[expr_cls_name]
    else:
        l.warning("Unsupported Soot expression %s.", expr_cls_name)
        expr_cls = SimSootExpr_Unsupported

    expr = expr_cls(expr, state)
    expr.process()
    return expr


from .arrayref import SimSootExpr_ArrayRef
from .binop import SimSootExpr_Binop
from .cast import SimSootExpr_Cast
from .condition import SimSootExpr_Condition
from .constants import (
    SimSootExpr_IntConstant,
    SimSootExpr_LongConstant,
    SimSootExpr_FloatConstant,
    SimSootExpr_DoubleConstant,
    SimSootExpr_StringConstant,
    SimSootExpr_ClassConstant,
    SimSootExpr_NullConstant,
)
from .instancefieldref import SimSootExpr_InstanceFieldRef
from .invoke import (
    SimSootExpr_SpecialInvoke,
    SimSootExpr_StaticInvoke,
    SimSootExpr_VirtualInvoke,
    SimSootExpr_InterfaceInvoke,
)
from .length import SimSootExpr_Length
from .local import SimSootExpr_Local
from .new import SimSootExpr_New
from .newArray import SimSootExpr_NewArray
from .newMultiArray import SimSootExpr_NewMultiArray
from .phi import SimSootExpr_Phi
from .staticfieldref import SimSootExpr_StaticFieldRef
from .thisref import SimSootExpr_ThisRef
from .paramref import SimSootExpr_ParamRef
from .unsupported import SimSootExpr_Unsupported
from .instanceOf import SimSootExpr_InstanceOf



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("dsuiext.dll")
prototypes = \
    {
        #
        'DsBrowseForContainerW': SimTypeFunction([SimTypePointer(SimStruct({"cbStruct": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "pszCaption": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pszTitle": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pszRoot": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pszPath": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cchPath": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pfnCallback": SimTypeBottom(label="BFFCALLBACK"), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwReturnFormat": SimTypeInt(signed=False, label="UInt32"), "pUserName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pPassword": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pszObjectClass": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cchObjectClass": SimTypeInt(signed=False, label="UInt32")}, name="DSBROWSEINFOW", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pInfo"]),
        #
        'DsBrowseForContainerA': SimTypeFunction([SimTypePointer(SimStruct({"cbStruct": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "pszCaption": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pszTitle": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pszRoot": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pszPath": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cchPath": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pfnCallback": SimTypeBottom(label="BFFCALLBACK"), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "dwReturnFormat": SimTypeInt(signed=False, label="UInt32"), "pUserName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pPassword": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pszObjectClass": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cchObjectClass": SimTypeInt(signed=False, label="UInt32")}, name="DSBROWSEINFOA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pInfo"]),
        #
        'DsGetIcon': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["dwFlags", "pszObjectClass", "cxImage", "cyImage"]),
        #
        'DsGetFriendlyClassName': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pszObjectClass", "pszBuffer", "cchBuffer"]),
    }

lib.set_prototypes(prototypes)



import angr
from angr.sim_options import MEMORY_CHUNK_INDIVIDUAL_READS
from angr.storage.memory_mixins.regioned_memory.abstract_address_descriptor import AbstractAddressDescriptor

import logging

l = logging.getLogger(name=__name__)


class strlen(angr.SimProcedure):
    # pylint:disable=arguments-differ
    max_null_index = None

    def run(self, s, wchar=False, maxlen=None):
        if wchar:
            null_seq = self.state.solver.BVV(0, 16)
            char_size = 2
        else:
            null_seq = self.state.solver.BVV(0, 8)
            char_size = 1

        max_symbolic_bytes = self.state.libc.buf_symbolic_bytes
        max_str_len = self.state.libc.max_str_len
        if maxlen:
            max_str_len = min(maxlen, max_str_len)

        chunk_size = None
        if MEMORY_CHUNK_INDIVIDUAL_READS in self.state.options:
            chunk_size = 1

        if self.state.mode == "static":
            self.max_null_index = 0

            # Make sure to convert s to ValueSet
            addr_desc: AbstractAddressDescriptor = self.state.memory._normalize_address(s)

            # size_t
            length = self.state.solver.ESI(self.arch.bits)
            for s_aw in self.state.memory._concretize_address_descriptor(addr_desc, None):
                s_ptr = s_aw.to_valueset(self.state)
                r, c, i = self.state.memory.find(
                    s,
                    null_seq,
                    max_str_len,
                    max_symbolic_bytes=max_symbolic_bytes,
                    chunk_size=chunk_size,
                    char_size=char_size,
                )

                self.max_null_index = max([self.max_null_index] + i)

                # Convert r to the same region as s
                r_desc = self.state.memory._normalize_address(r)
                r_aw_iter = self.state.memory._concretize_address_descriptor(
                    r_desc, None, target_region=next(iter(s_ptr._model_vsa.regions.keys()))
                )

                for r_aw in r_aw_iter:
                    r_ptr = r_aw.to_valueset(self.state)
                    length = length.union(r_ptr - s_ptr)

            return length

        else:
            search_len = max_str_len
            r, c, i = self.state.memory.find(
                s,
                null_seq,
                search_len,
                max_symbolic_bytes=max_symbolic_bytes,
                chunk_size=chunk_size,
                char_size=char_size,
            )

            # try doubling the search len and searching again
            s_new = s
            while c and all(con.is_false() for con in c):
                s_new += search_len
                search_len *= 2
                r, c, i = self.state.memory.find(
                    s_new,
                    null_seq,
                    search_len,
                    max_symbolic_bytes=max_symbolic_bytes,
                    chunk_size=chunk_size,
                    char_size=char_size,
                )
                # stop searching after some reasonable limit
                if search_len > 0x10000:
                    raise angr.SimMemoryLimitError("strlen hit limit of 0x10000")

            self.max_null_index = max(i)
            self.state.add_constraints(*c)
            result = r - s
            if result.depth > 3:
                rresult = self.state.solver.BVS("strlen", len(result), key=("api", "strlen"))
                self.state.add_constraints(result == rresult)
                result = rresult
            return result



import angr


class __ctype_toupper_loc(angr.SimProcedure):
    """
    Following is the description from linuxfoundation.org:

    The __ctype_toupper_loc() function shall return a pointer into an array
    of characters in the current locale that contains upper case equivalents
    for each character in the current character set. The array shall contain
    a total of 384 characters, and can be indexed with any signed or unsigned
    char (i.e. with an index value between -128 and 255). If the application
    is multithreaded, the array shall be local to the current thread.

    This interface is not in the source standard; it is only in the binary
    standard.
    """

    def run(self):
        table_ptr = self.state.libc.ctype_toupper_loc_table_ptr

        return table_ptr
